from Atomic import Atomic
from Atomic import util
from Atomic import info
from Atomic import update
from Atomic import verify
from Atomic import help as Help
from Atomic.mount import Mount
from Atomic.delete import Delete
from Atomic.tag import Tag
import os
import math
import shutil
import tempfile
import argparse
from Atomic import backendutils

ATOMIC_CONFIG = util.get_atomic_config()
storage = ATOMIC_CONFIG.get('default_storage', "docker")

def convert_size(size):
    if size > 0:
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size, 1000)))
        p = math.pow(1000, i)
        s = round(size/p, 2) # pylint: disable=round-builtin,old-division
        if s > 0:
            return '%s %s' % (s, size_name[i])
    return '0B'

def cli(subparser):
    # atomic images
    imagesp = subparser.add_parser("images",
                                   help=_("operate on images"))
    images_subparser = imagesp.add_subparsers(title='images subcommands',
                                              description="operate on images",
                                              help='additional help')

    # atomic images delete
    delete_parser = images_subparser.add_parser("delete",
                                                help=_("mark image for deletion"),
                                                epilog="Marks image. registry garbage-collection "
                                                "when invoked will recover used disk space")

    deletegroup = delete_parser.add_mutually_exclusive_group()
    deletegroup.add_argument("-f", "--force", default=False, dest="force",
                             action="store_true",
                             help=_("Force removal of local images, even if containers based on it exist.  Default is False"))

    deletegroup.add_argument("--remote", default=False, dest="remote",
                             action="store_true",
                             help=_("Delete image from remote repository"))

    delete_parser.add_argument("--storage", default=None, dest="storage",
                               help=_("Specify the storage from which to delete the image from. "
                                      "If not specified and there are images with the same name in "
                                      "different storages, you will be prompted to specify."))
    delete_parser.add_argument("-a", "--all", action='store_true',dest="all",
                               default=False, help=_("Delete all images"))
    delete_parser.add_argument("delete_targets", nargs='*',
                                    help=_("container image(s)"))
    delete_parser.set_defaults(_class=Delete, func='delete_image')

    if util.gomtree_available():
        generate_parser = images_subparser.add_parser("generate",
                                                  help=_("generate image 'manifests' if missing"))
        generate_parser.set_defaults(_class=Images, func='generate_validation_manifest')

        # Suppress storage flag: only docker images are supported for generate
        generate_parser.add_argument("--storage", default=storage, dest="storage",
                                     help=argparse.SUPPRESS)

    Help.cli(images_subparser)

    info.cli(images_subparser)

    # atomic images list
    list_parser = images_subparser.add_parser("list",
                                              help=_("list container images on your system"),
                                              epilog="atomic images by default will list all installed "
                                                     "container images on your system.")
    list_parser.set_defaults(_class=Images, func='display_all_image_info')

    list_parser.add_argument("-a", "--all", dest="all", default=False,
                             action="store_true",
                             help=_("Show all images, including intermediate images"))

    list_parser.add_argument("-f", "--filter", dest="filter", metavar="FILTER",
                             action="append",
                             help=_("Filter output based on VARIABLE=VALUE format"))

    list_parser.add_argument("-n", "--noheading", dest="heading", default=True,
                             action="store_false",
                             help=_("do not print heading when listing the images"))

    list_parser.add_argument("--no-trunc", dest="truncate", default=True,
                             action="store_false",
                             help=_("Don't truncate output"))

    list_parser.add_argument("-q", "--quiet", dest="quiet", default=False,
                             action="store_true",
                             help=_("Only display image IDs"))
    list_parser.add_argument("--json", action='store_true',dest="json", default=False,
                             help=_("print in a machine parseable form"))

    # atomic images tag
    tag_parser = images_subparser.add_parser("tag",
                                                help=_("Create a tag for the specified image"),
                                                epilog="Tag the specified image with a different name")
    tag_parser.add_argument("--storage", default=None, dest="storage",
                               help=_("Specify the storage to use for tagging the image. "
                                      "If not specified and there are images with the same name in "
                                      "different storages, you will be prompted to specify."))
    tag_parser.add_argument('src', metavar='SRC')
    tag_parser.add_argument('target', metavar='TARGET')

    tag_parser.set_defaults(_class=Tag, func='tag_image')

    prune_parser = images_subparser.add_parser("prune",
                                               help=_("delete unused 'dangling' images"),
                                               epilog="Using the prune command, "
                                                      "will free up disk space deleting unused "
                                                      "'dangling' images")
    prune_parser.set_defaults(_class=Delete, func='prune_images')

    update.cli(images_subparser)

    verify.cli(images_subparser)

    info.cli_version(images_subparser)

class Images(Atomic):

    FILTER_KEYWORDS = ["repo", "repository", "tag", "id", "image", "created", "size", "type", "is_dangling"]

    def __init__(self):
        super(Images, self).__init__()
        self.be_utils = backendutils.BackendUtils()

    def display_all_image_info(self):
        def get_col_lengths(_images):
            '''
            Determine the max length of the repository and tag names
            :param _images:
            :return: a set with len of repository and tag
            If there are no images, return 1, 1
            '''
            repo_tags = [y for x in _images if x.repotags for y in x.split_repotags]

            if repo_tags:
                return max([len(x[0]) for x in repo_tags]) + 2,\
                       max([len(x[1]) for x in repo_tags]) + 2
            else:
                return 1, 1

        if self.args.debug:
            util.write_out(str(self.args))
            self.be_utils.dump_backends()

        _images = self._get_images()

        if self.args.filter:
            self._check_filter_validity()

        if self.args.json:
            util.output_json(self.return_json(_images))
            return 0

        if len(_images) == 0:
            return

        _max_repo, _max_tag = get_col_lengths(_images)

        if self.args.truncate:
            _max_id = 14
        else:
            _max_id = 65
        col_out = "{0:2} {1:" + str(_max_repo) + "} {2:" + str(_max_tag) + \
                  "} {3:" + str(_max_id) + "} {4:18} {5:14} {6:10}"

        if self.args.heading and not self.args.quiet:
            util.write_out(col_out.format(" ",
                                          "REPOSITORY",
                                          "TAG",
                                          "IMAGE ID",
                                          "CREATED",
                                          "VIRTUAL SIZE",
                                          "TYPE"))
        for image in _images:
            _id = image.short_id if self.args.truncate else image.id
            if self.args.quiet:
                for _repo, _tag in image.split_repotags:
                    if self.args.filter and not self._filter_include_image(image, _repo, _tag):
                        continue
                    util.write_out(_id)

            else:
                indicator = ""
                if image.is_dangling:
                    indicator += "*"
                elif image.used:
                    indicator += ">"
                if image.vulnerable:
                    space = " " if len(indicator) < 1 else ""
                    if util.is_python2:
                        indicator = indicator + self.skull + space
                    else:
                        indicator = indicator + str(self.skull, "utf-8") + space

                for _repo, _tag in image.split_repotags:
                    if self.args.filter and not self._filter_include_image(image, _repo, _tag):
                        continue
                    util.write_out(col_out.format(indicator, _repo or "<none>", _tag or "<none>", _id, image.created[0:16],
                                                  image.virtual_size, image.backend.backend))
        return

    def _get_images(self):
        _images = self.be_utils.get_images(get_all=self.args.all)

        self._mark_used(_images)
        self._mark_vulnerable(_images)

        return _images

    def images(self):
        return self.return_json(self._get_images())


    def generate_validation_manifest(self):
        """
        Generates a gomtree validation manifest for a non-system image and stores it in
        ATOMIC_VAR_LIB
        :param:
        :return: None
        """
        _images = self.get_images(get_all=True)
        for image in _images:
            atomic_var_lib = util.ATOMIC_VAR_LIB
            if not image["RepoTags"]:
                continue
            iid = image["RepoTags"][0]
            if image["ImageType"] == "system":
                continue
            if iid == "<none>:<none>" or iid == "<none>":
                continue
            if os.path.exists(os.path.join(atomic_var_lib, "gomtree-manifests/%s.mtree" % iid)):
                continue
            manifestname = os.path.join(atomic_var_lib, "gomtree-manifests/%s.mtree" % iid)
            dname = os.path.dirname(manifestname)
            if not os.path.exists(dname):
                os.makedirs(dname)
            tmpdir = tempfile.mkdtemp()
            m = Mount()
            m.args = []
            m.image = iid
            m.mountpoint = tmpdir
            m.storage = self.args.storage
            m.mount()
            r = util.generate_validation_manifest(img_rootfs=tmpdir, keywords="type,uid,gid,mode,size,sha256digest")
            m.unmount()
            with open(manifestname,"wb",0) as f:
                f.write(r.stdout)
            shutil.rmtree(tmpdir)

    def _check_filter_validity(self):
        for i in self.args.filter:
            try:
                var, _ = str(i).split("=")
            except ValueError:
                raise ValueError("The filter {} is not formatted correctly.  It should be VAR=VALUE".format(i))

            var = var.lower()
            if var not in self.FILTER_KEYWORDS:
                raise ValueError("The filter {} is not valid.  "
                                 "Please choose from {}".format(var, self.FILTER_KEYWORDS))


    def _filter_include_image(self, image_obj, repo, tag):
        for i in self.args.filter:
            var, value = str(i).split("=")
            var = var.lower()
            if var == "repository":
                var = "repo"

            if var == "image":
                var = "id"

            if var == "type":
                var = "str_backend"

            if var == "repo":
                if not value.lower() in repo:
                    return False

            elif var == "tag":
                if not value.lower() in tag:
                    return False

            elif not hasattr(image_obj, var):
                return False

            elif hasattr(image_obj, var) and value.lower() not in str(getattr(image_obj, var)).lower():
                return False

        return True

    def _mark_used(self, images):
        assert isinstance(images, list)
        all_containers = [x.image for x in self.be_utils.get_containers()]
        for image in images:
            if image.id in all_containers:
                image.used = True

    def _mark_vulnerable(self, images):
        assert isinstance(images, list)
        vulnerable_uuids = self.get_vulnerable_ids()
        for image in images:
            if image.id in vulnerable_uuids:
                image.vulnerable = True

    def return_json(self, images):
        all_image_info = []
        all_vuln_info = self.get_all_vulnerable_info()
        keys = ['is_dangling', 'used', 'vulnerable', 'id', 'type', 'created', 'virtual_size', 'str_backend']
        keys_rename = {"str_backend" : "backend"}
        for img_obj in images:
            if not img_obj.repotags:
                continue

            for _repo, _tag in img_obj.split_repotags:
                if self.args.filter and not self._filter_include_image(img_obj, _repo, _tag):
                    continue

                img_dict = dict()
                img_dict['repo'], img_dict['tag'] = _repo, _tag
                for key in keys:
                    key_name = key if key not in keys_rename else keys_rename[key]
                    img_dict[key_name] = getattr(img_obj, key, None)
                img_dict['vuln_info'] = \
                    dict() if not img_obj.vulnerable else all_vuln_info.get(img_obj.id, None) # pylint: disable=no-member
                all_image_info.append(img_dict)

        return all_image_info





#!/usr/bin/env python

# Author: Dan Walsh <dwalsh@redhat.com>
from distutils.core import setup
import Atomic as _Atomic

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="atomic", scripts=["atomic", "atomic_dbus.py"],
    version=_Atomic.__version__,
    author=_Atomic.__author__,
    author_email=_Atomic.__author_email__,
    packages=["Atomic", "Atomic/backends", "Atomic/objects"],
    data_files=[('/etc/dbus-1/system.d/', ["org.atomic.conf"]),
                ('/usr/share/dbus-1/system-services', ["org.atomic.service"]),
                ('/usr/share/polkit-1/actions/', ["org.atomic.policy"]),
                ("/usr/share/bash-completion/completions/",
                 ["bash/atomic"])]
)



#pylint: skip-file
import unittest
from Atomic.backendutils import BackendUtils
from Atomic.backends._docker import DockerBackend
from Atomic.backends._ostree import OSTreeBackend
from Atomic.syscontainers import SystemContainers
from Atomic.info import Info
from Atomic.images import Images
from Atomic.delete import Delete

no_mock = True
try:
    from unittest.mock import MagicMock, patch
    no_mock = False
except ImportError:
    try:
        from mock import MagicMock, patch
        no_mock = False
    except ImportError:
        # Mock is already set to False
        pass

docker_images = [{'Id': '7968321274dc6b6171697c33df7815310468e694ac5be0ec03ff053bb135e768', 'Size': 1109996, 'ParentId': '', 'Created': 1484345634, 'RepoTags': ['docker.io/busybox:latest'], 'VirtualSize': 1109996, 'RepoDigests': ['docker.io/busybox@sha256:817a12c32a39bbe394944ba49de563e085f1d3c5266eb8e9723256bc4448680e'], 'Labels': {}}, {'Id': '88e169ea8f46ff0d0df784b1b254a15ecfaf045aee1856dca1ec242fdd231ddd', 'Size': 3979756, 'ParentId': '', 'Created': 1482862645, 'RepoTags': ['docker.io/alpine:latest'], 'VirtualSize': 3979756, 'RepoDigests': ['docker.io/alpine@sha256:dfbd4a3a8ebca874ebd2474f044a0b33600d4523d03b0df76e5c5986cb02d7e8'], 'Labels': None}]
ostree_images = [{'Id': '5040bd2983909aa8896b9932438c3f1479d25ae837a5f6220242a264d0221f2d', 'Labels': {}, 'ImageType': 'system', 'Version': '<none>', 'ImageId': '5040bd2983909aa8896b9932438c3f1479d25ae837a5f6220242a264d0221f2d', 'RepoTags': ['<none>'], 'OSTree-rev': 'fe2619a941bff9f323d1f5eef7fc3e05f2c82d421c4ba383ec646c888231752b', 'Created': 1486143371, 'Names': []}, {'Id': 'e5599115b6a67e08278d176b05a3defb30e5564f5be6d73264ec560b484514a2', 'Labels': {}, 'ImageType': 'system', 'Version': 'debian:latest', 'ImageId': 'e5599115b6a67e08278d176b05a3defb30e5564f5be6d73264ec560b484514a2', 'RepoTags': ['debian:latest'], 'OSTree-rev': '8d4551074942f65080a613d68a21f131fdebc0bf8ce437dd676bb4dbfa50fa08', 'Created': 1486143371, 'Names': []}, {'Id': '45a2e645736c4c66ef34acce2407ded21f7a9b231199d3b92d6c9776df264729', 'Labels': {}, 'ImageType': 'system', 'Version': '<none>', 'ImageId': '45a2e645736c4c66ef34acce2407ded21f7a9b231199d3b92d6c9776df264729', 'RepoTags': ['<none>'], 'OSTree-rev': 'a2512de2d19ab0916f9acce1ebd731203a336c366a1d002126c198ceff280375', 'Created': 1486143486, 'Names': []}, {'Id': '67591570dd29de0e124ee89d50458b098dbd83b12d73e5fdaf8b4dcbd4ea50f8', 'Labels': {}, 'ImageType': 'system', 'Version': 'centos:latest', 'ImageId': '67591570dd29de0e124ee89d50458b098dbd83b12d73e5fdaf8b4dcbd4ea50f8', 'RepoTags': ['centos:latest'], 'OSTree-rev': '1ce1d5393d4c1d4e04d70576be95ca598b26dff14aa9deacc92e06df5f71c470', 'Created': 1486143486, 'Names': []}]


@unittest.skipIf(no_mock, "Mock not found")
class TestInfo(unittest.TestCase):
    class Args():
        def __init__(self):
            self.storage = None
            self.force = False
            self.all = False
            self.assumeyes = True
            self.debug = False
            self.delete_targets = []
            self.remote = False

    def test_delete_no_images(self):
        with patch('Atomic.backendutils.BackendUtils.get_images') as mockobj:
            args = self.Args()
            args.all = True
            del_ = Delete()
            del_.set_args(args)
            mockobj.return_value([])
            with self.assertRaises(ValueError):
                del_.delete_image()

    def test_delete_no_images_ostree(self):
        with patch('Atomic.backends._ostree.OSTreeBackend.get_images') as mockobj:
            args = self.Args()
            args.all = True
            args.storage = 'ostree'
            del_ = Delete()
            del_.set_args(args)
            mockobj.return_value([])
            with self.assertRaises(ValueError):
                del_.delete_image()

    def test_delete_no_images_docker(self):
        with patch('Atomic.backends._docker.DockerBackend.get_images') as mockobj:
            args = self.Args()
            args.all = True
            args.storage = 'docker'
            del_ = Delete()
            del_.set_args(args)
            mockobj.return_value([])
            with self.assertRaises(ValueError):
                del_.delete_image()

    def test_delete_all_and_images(self):
        args = self.Args()
        args.all = True
        args.delete_targets = 'foobar'
        del_ = Delete()
        del_.set_args(args)
        with self.assertRaises(ValueError):
            del_.delete_image()

    def test_delete_not_all_and_not_images(self):
        args = self.Args()
        args.delete_targets = []
        del_ = Delete()
        del_.set_args(args)
        with self.assertRaises(ValueError):
            del_.delete_image()


    def test_delete_all_docker(self):
        with patch('Atomic.backends._docker.DockerBackend.delete_image') as deleteobj:
            with patch('Atomic.backends._docker.DockerBackend._get_images') as imageobj:
                args = self.Args()
                args.all = True
                args.storage = 'docker'
                del_ = Delete()
                del_.set_args(args)
                deleteobj.return_value = None
                imageobj.return_value = docker_images
                self.assertEqual(del_.delete_image(), 0)

    def test_delete_all_ostree(self):
        with patch('Atomic.backends._ostree.OSTreeBackend.delete_image') as deleteobj:
            with patch('Atomic.syscontainers.SystemContainers.get_system_images') as imageobj:
                args = self.Args()
                args.all = True
                args.storage = 'ostree'
                del_ = Delete()
                del_.set_args(args)
                deleteobj.return_value = None
                imageobj.return_value = ostree_images
                self.assertEqual(del_.delete_image(), 0)



#pylint: skip-file

import unittest
from Atomic.util import is_backend_available, load_registries_from_yaml, get_registries
import json
import subprocess

no_mock = True
try:
    from unittest.mock import MagicMock, patch
    no_mock = False
except ImportError:
    try:
        from mock import MagicMock, patch
        no_mock = False
    except ImportError:
        # Mock is already set to False
        pass


@unittest.skipIf(no_mock, "Mock not found")
class TestRegistriesFromYAML(unittest.TestCase):

    def compare_list_of_dicts(self, results, answer):
        if not isinstance(results, list) or not isinstance(answer, list):
            raise AssertionError("Results must always be of type list")
        if len(results) != len(answer):
            raise AssertionError("Length of the lists differ.")
        for registry in results:
            self.assertDictEqual(registry, next(item for item in answer if item["hostname"] == registry['hostname']))

    def test_no_registries(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads("{}")
                results = get_registries()
        answer = [{'name': 'docker.io', 'hostname': 'registry-1.docker.io', 'search': True, 'secure': True}]
        self.compare_list_of_dicts(results, answer)

    def test_block_dockerio(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads('{"block_registries": ["docker.io"]}')
                results = get_registries()
        answer = []
        self.compare_list_of_dicts(results, answer)


    def test_duplicate_dockerio(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads('{"registries": ["docker.io"]}')
                results = get_registries()
        answer = [{'secure': True, 'hostname': 'docker.io', 'name': 'docker.io', 'search': True}]
        self.compare_list_of_dicts(results, answer)

    def test_all(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads('{"registries": ["one.com", "two.com"], "insecure_registries": ["three.com"], "block_registries": []}')
                results = get_registries()
        answer = [{'secure': True, 'search': True, 'name': 'one.com', 'hostname': 'one.com'}, {'secure': True, 'search': True, 'name': 'two.com', 'hostname': 'two.com'}, {'secure': True, 'search': True, 'name': 'three.com', 'hostname': 'three.com'}, {'secure': True, 'search': True, 'name': 'docker.io', 'hostname': 'registry-1.docker.io'}]
        self.compare_list_of_dicts(results, answer)

    def test_duplicate_in_secure_and_insecure(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads('{"registries": ["one.com", "two.com"], "insecure_registries": ["two.com"], "block_registries": []}')
                self.assertRaises(ValueError, get_registries)

    def test_duplicate_in_registries(self):
        with patch('Atomic.util.registries_tool_path') as reg_path:
            reg_path.return_value = "/usr/libexec/registries"
            with patch('Atomic.util.load_registries_from_yaml') as mockobj:
                mockobj.return_value = json.loads('{"registries": ["one.com", "two.com", "one.com"], "insecure_registries": ["three.com"], "block_registries": []}')
                results = get_registries()
        answer = [{'secure': True, 'search': True, 'name': 'one.com', 'hostname': 'one.com'}, {'secure': True, 'search': True, 'name': 'two.com', 'hostname': 'two.com'}, {'secure': True, 'search': True, 'name': 'three.com', 'hostname': 'three.com'}, {'secure': True, 'search': True, 'name': 'docker.io', 'hostname': 'registry-1.docker.io'}]
        self.compare_list_of_dicts(results, answer)

if __name__ == '__main__':
    unittest.main()
